package thesis.algorithms.evolutionary.bivariate.dependecy_graphs;

import java.util.HashSet;
import java.util.Random;

import thesis.algorithms.evolutionary.bivariate.Edge;
import thesis.algorithms.evolutionary.bivariate.probability_matrix.IProbabilityMatrix;
import thesis.util.Chromosome;

public class BMDADependencyGraph implements IDependencyGraph {

	private HashSet<Edge> edges;
	private HashSet<Integer> verticesA;
	private HashSet<Integer> verticesR;
	private HashSet<Integer> verticesV;
	private HashSet<Integer> verticesK;
	private int bitsNo;

	public BMDADependencyGraph(int bitsNo) {

		this.bitsNo = bitsNo;

		edges = new HashSet<Edge>();
		verticesA = new HashSet<Integer>();
		verticesR = new HashSet<Integer>();
		verticesV = new HashSet<Integer>();
		verticesK = new HashSet<Integer>();
	}

	@Override
	public void constructDependencyGraph(double[][] chiSquare) {
		
		fillVertices(verticesA);
		fillVertices(verticesV);

		verticesR.clear();
		verticesK.clear();

		boolean chooseNewVertex = true;
		Integer vertex = null;

		while (!verticesA.isEmpty()) {

			if (chooseNewVertex) {

				vertex = verticesA.iterator().next();
				verticesR.add(vertex);
			}

			verticesA.remove(vertex);

			double maximumDependecy = -1;
			Integer vertexFromA = -1, vertexFromVWithoutA = -1;

			for (Integer i = 0; i < bitsNo; ++i) {
				for (Integer j = i + 1; j < bitsNo; ++j) {

					if (!verticesA.contains(i)) {
						continue;
					}

					if (verticesA.contains(j)) {
						continue;
					}

					if (chiSquare[i][j] > 3.84) {
						if (chiSquare[i][j] > maximumDependecy) {
							maximumDependecy = chiSquare[i][j];
							vertexFromA = i;
							vertexFromVWithoutA = j;
						}
					}
				}
			}

			if (maximumDependecy == -1) {
				chooseNewVertex = true;
				continue;
			}

			chooseNewVertex = false;
			vertex = vertexFromA;
			edges.add(new Edge(vertexFromA, vertexFromVWithoutA, bitsNo));
		}
	}

	private void fillVertices(HashSet<Integer> vertices) {

		vertices.clear();
		for (int i = 0; i < bitsNo; ++i) {
			vertices.add(i);
		}
	}

	@Override
	public Chromosome generateNewIndividual(IProbabilityMatrix probabilityMatrix) {

		Random randomGenerator = new Random();
		Chromosome newIndividual = new Chromosome(bitsNo);
		fillVertices(verticesK);

		for (Integer vertex : verticesR) {

			// setting true values if needed (false is default)
			if (randomGenerator.nextDouble() < probabilityMatrix
					.getProbability(vertex)) {

				newIndividual.setBit(vertex, true);
			}
			
			verticesK.remove(vertex);
		}

		while (!verticesK.isEmpty()) {

			for (Edge edge : edges) {

				if (verticesK.contains(edge.getFirstVertex())
						&& verticesK.contains(edge.getSecondVertex())) {
					continue;
				}

				if (!verticesK.contains(edge.getFirstVertex())
						&& !verticesK.contains(edge.getSecondVertex())) {
					continue;
				}

				int generatingBitIndex;
				int hypothesisBitIndex;
				
				if (verticesK.contains(edge.getFirstVertex())) {

					generatingBitIndex = edge.getFirstVertex();
					hypothesisBitIndex = edge.getSecondVertex();
					verticesK.remove(edge.getFirstVertex());
				}else{
					
					generatingBitIndex = edge.getSecondVertex();
					hypothesisBitIndex = edge.getFirstVertex();
					verticesK.remove(edge.getSecondVertex());
				}

				boolean generatedBit = generateBit(generatingBitIndex,
						hypothesisBitIndex, probabilityMatrix);
				newIndividual.setBit(generatingBitIndex, generatedBit);
			}
		}

		return newIndividual;
	}

	private boolean generateBit(int generatingBit, int hypothesisBit,
			IProbabilityMatrix probabilityMatrix) {

		double pJ = probabilityMatrix.getProbability(hypothesisBit);
		double pIJ = probabilityMatrix.getJointProbability(generatingBit,
				hypothesisBit);

		double probabilityI = pIJ / pJ;

		Random randomGenerator = new Random();
		if (randomGenerator.nextDouble() < probabilityI) {
			return true;
		}
		return false;
	}

	@Override
	public Chromosome[] generateNewIndiviudals(
			IProbabilityMatrix probabilityMatrix, int size) {

		Chromosome[] solutions = new Chromosome[size];

		for (int i = 0; i < size; ++i) {
			solutions[i] = generateNewIndividual(probabilityMatrix);
		}
		return solutions;
	}

}
